#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import inspect
import os
from configparser import ConfigParser

from fenrirscreenreader.core import applicationManager
from fenrirscreenreader.core import attributeManager
from fenrirscreenreader.core import barrierManager
from fenrirscreenreader.core import commandManager
from fenrirscreenreader.core import cursorManager
from fenrirscreenreader.core import debug
from fenrirscreenreader.core import debugManager
from fenrirscreenreader.core import diffReviewManager
from fenrirscreenreader.core import environment
from fenrirscreenreader.core import eventManager
from fenrirscreenreader.core import helpManager
from fenrirscreenreader.core import inputManager
from fenrirscreenreader.core import memoryManager
from fenrirscreenreader.core import outputManager
from fenrirscreenreader.core import processManager
from fenrirscreenreader.core import punctuationManager
from fenrirscreenreader.core import quickMenuManager
from fenrirscreenreader.core import readAllManager
from fenrirscreenreader.core import remoteManager
from fenrirscreenreader.core import sayAllManager
from fenrirscreenreader.core import screenManager
from fenrirscreenreader.core import tableManager
from fenrirscreenreader.core import textManager
from fenrirscreenreader.core import vmenuManager
from fenrirscreenreader.core.settingsData import settings_data
from fenrirscreenreader.utils import module_utils

currentdir = os.path.dirname(
    os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
)
fenrir_path = os.path.dirname(currentdir)


class SettingsManager:
    system_settings_root = "/etc/fenrirscreenreader/"
    system_settings_file = "/etc/fenrirscreenreader/settings/settings.conf"
    user_settings_file = (
        "~/.local/share/stormux/fenrirscreenreader/settings/settings.conf"
    )
    user_resource_root = "~/.local/stormux/fenrir/"
    system_sound_roots = [
        "/usr/share/sounds/fenrir/",
        "/usr/share/sounds/fenrirscreenreader/",
    ]

    def __init__(self):
        self.settings = settings_data
        self.settingArgDict = {}
        self.bindingsBackup = None
        self.settings_file = ""
        self.save_settings_path = ""

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_binding_backup(self):
        return self.bindingsBackup.copy()

    def load_sound_icons(self, soundIconPath, environment=None):
        # Use passed environment or fall back to self.env
        env = environment if environment is not None else self.env
        try:
            with open(soundIconPath + "/soundicons.conf", "r") as siConfig:
                while True:
                    line = siConfig.readline()
                    if not line:
                        break
                    line = line.replace("\n", "")
                    if line.replace(" ", "") == "":
                        continue
                    if line.replace(" ", "").startswith("#"):
                        continue
                    if line.count("=") != 1:
                        continue
                    values = line.split("=")
                    sound_icon = values[0].upper()
                    values[1] = values[1].replace("'", "")
                    values[1] = values[1].replace('"', "")
                    sound_icon_file = ""
                    if os.path.exists(values[1]):
                        sound_icon_file = values[1]
                    else:
                        if not soundIconPath.endswith("/"):
                            soundIconPath += "/"
                        if os.path.exists(soundIconPath + values[1]):
                            sound_icon_file = soundIconPath + values[1]
                    env["soundIcons"][sound_icon] = sound_icon_file
                    env["runtime"]["DebugManager"].write_debug_out(
                        "SoundIcon: " + sound_icon + "." + sound_icon_file,
                        debug.DebugLevel.INFO,
                        on_any_level=True,
                    )
        except (IOError, OSError) as e:
            env["runtime"]["DebugManager"].write_debug_out(
                "load_sound_icons: failed to load sound icons from "
                + soundIconPath
                + ". Error: "
                + str(e),
                debug.DebugLevel.ERROR,
            )

    def get_settings_file(self):
        return self.settings_file

    def set_settings_file(self, settings_file):
        if os.path.exists(settings_file) and not os.access(
            settings_file, os.R_OK
        ):
            return
        self.settings_file = settings_file

    def get_user_settings_file(self):
        return os.path.expanduser(self.user_settings_file)

    def get_system_settings_file(self):
        return self.system_settings_file

    def is_user_settings_mode(self):
        return os.geteuid() != 0

    def get_default_save_settings_file(self):
        if self.is_user_settings_mode():
            return self.get_user_settings_file()
        return self.get_system_settings_file()

    def get_bundled_settings_root(self):
        return os.path.abspath(os.path.join(fenrir_path, "../../config/")) + "/"

    def get_user_resource_root(self):
        return os.path.expanduser(self.user_resource_root)

    def get_user_script_path(self):
        return self.get_user_resource_root()

    def get_user_sound_root(self):
        return os.path.join(self.get_user_resource_root(), "sounds/")

    def get_resource_settings_root(self):
        if os.path.exists(self.system_settings_root):
            return self.system_settings_root
        bundled_settings_root = self.get_bundled_settings_root()
        if os.path.exists(bundled_settings_root):
            return bundled_settings_root
        return ""

    def get_sound_theme_roots(self):
        candidate_roots = [
            self.get_user_sound_root(),
            *self.system_sound_roots,
            os.path.join(self.get_bundled_settings_root(), "sound/"),
        ]
        sound_roots = []
        seen_roots = set()
        for root in candidate_roots:
            normalized_root = os.path.abspath(os.path.expanduser(root))
            if normalized_root in seen_roots:
                continue
            sound_roots.append(normalized_root + "/")
            seen_roots.add(normalized_root)
        return sound_roots

    def resolve_sound_theme_path(self, theme):
        if not theme:
            return ""
        theme = os.path.expanduser(theme)
        if os.path.exists(os.path.join(theme, "soundicons.conf")):
            return theme
        for sound_root in self.get_sound_theme_roots():
            theme_path = os.path.join(sound_root, theme)
            if os.path.exists(os.path.join(theme_path, "soundicons.conf")):
                return theme_path
        return ""

    def load_keyboard_layout(self, environment):
        settings_root = self.get_resource_settings_root()
        keyboard_layout = self.get_setting("keyboard", "keyboard_layout")
        if os.path.exists(keyboard_layout):
            environment["runtime"]["InputManager"].load_shortcuts(
                keyboard_layout
            )
            return

        layout_path = os.path.join(settings_root, "keyboard", keyboard_layout)
        if os.path.exists(layout_path):
            self.set_setting("keyboard", "keyboard_layout", layout_path)
            environment["runtime"]["InputManager"].load_shortcuts(layout_path)
            return

        layout_path = os.path.join(
            settings_root, "keyboard", keyboard_layout + ".conf"
        )
        if os.path.exists(layout_path):
            self.set_setting("keyboard", "keyboard_layout", layout_path)
            environment["runtime"]["InputManager"].load_shortcuts(layout_path)

    def resolve_settings_file(self, requested_settings_file=None):
        if requested_settings_file and os.path.exists(requested_settings_file):
            return requested_settings_file

        if self.is_user_settings_mode() and os.path.exists(
            self.get_user_settings_file()
        ):
            return self.get_user_settings_file()

        if os.path.exists(self.get_system_settings_file()):
            return self.get_system_settings_file()

        bundled_settings_file = os.path.join(
            self.get_bundled_settings_root(), "settings/settings.conf"
        )
        if os.path.exists(bundled_settings_file):
            return bundled_settings_file

        return ""

    def load_settings(self, setting_config_path):
        if not os.path.exists(setting_config_path):
            return False
        if not os.access(setting_config_path, os.R_OK):
            return False
        self.env["settings"] = ConfigParser()
        try:
            self.env["settings"].read(setting_config_path)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "settings_manager load_settings: Error reading config file: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
            return False
        self.set_settings_file(setting_config_path)
        return True

    def save_settings(self, setting_config_path=None):
        if setting_config_path is None:
            if self.save_settings_path:
                setting_config_path = self.save_settings_path
            else:
                setting_config_path = self.get_default_save_settings_file()

        # Ensure directory exists for user-local settings
        if setting_config_path and setting_config_path.startswith(
            os.path.expanduser("~/.local/")
        ):
            config_dir = os.path.dirname(setting_config_path)
            if not os.path.exists(config_dir):
                try:
                    os.makedirs(config_dir, mode=0o755, exist_ok=True)
                except Exception as e:
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "save_settings: failed to create directory "
                        + config_dir
                        + ": "
                        + str(e),
                        debug.DebugLevel.ERROR,
                    )
        # set opt dict here
        # save file
        try:
            # print('file: ',setting_config_path)
            for section, settings in self.settingArgDict.items():
                for setting, value in settings.items():
                    # print(section, setting, value)
                    self.env["settings"].set(section, setting, value)
            # print('full',self.env['settings'])

            config_file = open(setting_config_path, "w")
            self.env["settings"].write(config_file)
            config_file.close()
            os.chmod(setting_config_path, 0o644)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "save_settings: save settingsfile:"
                + setting_config_path
                + "failed. Error:"
                + str(e),
                debug.DebugLevel.ERROR,
            )

    def set_setting(self, section, setting, value):
        self.set_option_arg_dict(section, setting, value)
        # self.env['settings'].set(section, setting, value)

    def get_setting(self, section, setting):
        value = ""
        try:
            value = self.settingArgDict[section][setting]
            return value
        except Exception as e:
            pass
        try:
            value = self.env["settings"].get(section, setting)
        except Exception as e:
            value = str(self.settings[section][setting])
        return value

    def get_setting_as_int(self, section, setting):
        value = 0
        try:
            value = int(self.settingArgDict[section][setting])
            return value
        except Exception as e:
            pass
        try:
            value = self.env["settings"].getint(section, setting)
        except Exception as e:
            value = self.settings[section][setting]
        return value

    def get_setting_as_float(self, section, setting):
        value = 0.0
        try:
            value = float(self.settingArgDict[section][setting])
            return value
        except Exception as e:
            pass
        try:
            value = self.env["settings"].getfloat(section, setting)
        except Exception as e:
            value = self.settings[section][setting]
        return value

    def get_setting_as_bool(self, section, setting):
        value = False
        try:
            value = self.settingArgDict[section][setting].upper() in [
                "1",
                "YES",
                "JA",
                "TRUE",
            ]
            return value
        except Exception as e:
            pass
        try:
            value = self.env["settings"].getboolean(section, setting)
        except Exception as e:
            value = self.settings[section][setting]
        return value

    def load_driver(self, driverName, driverType):
        # Map runtime keys to actual directory names
        driver_dir_map = {
            "InputDriver": "inputDriver",
            "ScreenDriver": "screenDriver",
            "SpeechDriver": "speechDriver",
            "SoundDriver": "soundDriver",
            "RemoteDriver": "remoteDriver",
        }
        driver_dir = driver_dir_map.get(driverType, driverType)

        try:
            if self.env["runtime"][driverType] is not None:
                self.env["runtime"][driverType].shutdown()
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "settings_manager load_driver: Error shutting down driver: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
        try:
            driver_mod = module_utils.import_module(
                driverName,
                fenrir_path + "/" + driver_dir + "/" + driverName + ".py",
            )
            self.env["runtime"][driverType] = driver_mod.driver()
            self.env["runtime"][driverType].initialize(self.env)
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Loading Driver " + driverType + " (" + driverName + ") OK",
                debug.DebugLevel.INFO,
                on_any_level=True,
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Loading Driver "
                + driverType
                + " ("
                + driverName
                + ") FAILED:"
                + str(e),
                debug.DebugLevel.ERROR,
            )
            try:
                driver_mod = module_utils.import_module(
                    driverName,
                    fenrir_path + "/" + driver_dir + "/dummyDriver.py",
                )
                self.env["runtime"][driverType] = driver_mod.driver()
                self.env["runtime"][driverType].initialize(self.env)
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "(fallback) Loading Driver "
                    + driverType
                    + " (dummyDriver) FAILED:"
                    + str(e),
                    debug.DebugLevel.ERROR,
                )

    def shutdown_driver(self, driverType):
        try:
            self.env["runtime"][driverType].shutdown()
        except Exception as e:
            pass
        del self.env["runtime"][driverType]

    def set_fenrir_keys(self, keys):
        keys = keys.upper()
        key_list = keys.split(",")
        for key in key_list:
            if key not in self.env["input"]["fenrir_key"]:
                self.env["input"]["fenrir_key"].append(key)

    def set_script_keys(self, keys):
        keys = keys.upper()
        key_list = keys.split(",")
        for key in key_list:
            if key not in self.env["input"]["script_key"]:
                self.env["input"]["script_key"].append(key)

    def reset_setting_arg_dict(self):
        self.settingArgDict = {}
        self.env["runtime"]["OutputManager"].reset_SpeechDriver()

    def set_option_arg_dict(self, section, setting, value):
        # section = section.lower()
        # setting = setting.lower()
        try:
            e = self.settingArgDict[section]
        except KeyError:
            self.settingArgDict[section] = {}
        try:
            t = self.settings[section][setting]
        except Exception as e:
            print(section, setting, "not found")
            return
        try:
            v = value  # Initialize v with the original value
            if isinstance(self.settings[section][setting], str):
                v = str(value)
            elif isinstance(self.settings[section][setting], bool):
                if value not in ["True", "False"]:
                    raise ValueError(
                        "could not convert string to bool: " + value
                    )
                v = value == "True"
            elif isinstance(self.settings[section][setting], int):
                v = int(value)
            elif isinstance(self.settings[section][setting], float):
                v = float(value)

            # Content validation for critical settings
            self._validate_setting_value(section, setting, v)

            self.settingArgDict[section][setting] = str(value)
        except Exception as e:
            print(
                "settings_manager:set_option_arg_dict:Datatype missmatch: "
                + section
                + "#"
                + setting
                + "="
                + str(value)
                + " Error:"
                + str(e)
            )
            # self.env['runtime']['DebugManager'].write_debug_out('settings_manager:set_option_arg_dict:Datatype
            # missmatch: '+ section + '#' + setting + '=' +  value + ' Error:'
            # +  str(e), debug.DebugLevel.ERROR)
            return

    def _validate_setting_value(self, section, setting, value):
        """Validate setting values for critical screen reader functionality.
        Only validates settings that could cause crashes or accessibility issues.
        Invalid values raise ValueError which is caught by the calling method.
        """

        # Speech settings validation - critical for accessibility
        if section == "speech":
            if setting == "rate":
                if not (0.0 <= value <= 3.0):
                    raise ValueError(
                        f"Speech rate must be between 0.0 and 3.0, got {value}"
                    )
            elif setting == "pitch":
                if not (0.0 <= value <= 2.0):
                    raise ValueError(
                        f"Speech pitch must be between 0.0 and 2.0, got {value}"
                    )
            elif setting == "volume":
                if not (0.0 <= value <= 1.5):
                    raise ValueError(
                        f"Speech volume must be between 0.0 and 1.5, got {value}"
                    )
            elif setting == "driver":
                valid_drivers = [
                    "speechdDriver",
                    "genericDriver",
                    "dummyDriver",
                ]
                if value not in valid_drivers:
                    raise ValueError(
                        f"Invalid speech driver: {value}. Valid options: {valid_drivers}"
                    )

        # Sound settings validation
        elif section == "sound":
            if setting == "volume":
                if not (0.0 <= value <= 1.5):
                    raise ValueError(
                        f"Sound volume must be between 0.0 and 1.5, got {value}"
                    )
            elif setting == "driver":
                valid_drivers = [
                    "genericDriver",
                    "gstreamerDriver",
                    "dummyDriver",
                ]
                if value not in valid_drivers:
                    raise ValueError(
                        f"Invalid sound driver: {value}. Valid options: {valid_drivers}"
                    )

        # Screen settings validation
        elif section == "screen":
            if setting == "driver":
                valid_drivers = ["vcsaDriver", "ptyDriver", "dummyDriver"]
                if value not in valid_drivers:
                    raise ValueError(
                        f"Invalid screen driver: {value}. Valid options: {valid_drivers}"
                    )

        # Input settings validation
        elif section == "keyboard":
            if setting == "driver":
                valid_drivers = [
                    "evdevDriver",
                    "x11Driver",
                    "dummyDriver",
                ]
                if value not in valid_drivers:
                    raise ValueError(
                        f"Invalid input driver: {value}. Valid options: {valid_drivers}"
                    )

        # General settings validation
        elif section == "general":
            if setting == "debug_level":
                if not (0 <= value <= 3):
                    raise ValueError(
                        f"Debug level must be between 0 and 3, got {value}"
                    )

    def parse_setting_args(self, settingArgs):
        if settingArgs is None:
            return
        for optionElem in settingArgs.split(";"):
            setting_val_list = []
            section_option_list = []
            section = ""
            option = ""
            value = ""
            setting_val_list = optionElem.split("=", 1)
            if len(setting_val_list) != 2:
                continue
            if "#" in setting_val_list[0]:
                section_option_list = setting_val_list[0].split("#", 1)
            elif "." in setting_val_list[0]:
                section_option_list = setting_val_list[0].split(".", 1)
            elif "," in setting_val_list[0]:
                section_option_list = setting_val_list[0].split(",", 1)
            elif "!" in setting_val_list[0]:
                section_option_list = setting_val_list[0].split("!", 1)
            else:
                continue
            if len(section_option_list) != 2:
                continue

            section = str(section_option_list[0])
            option = str(section_option_list[1])
            value = str(setting_val_list[1])
            self.set_option_arg_dict(section, option, value)

    def init_fenrir_config(
        self, cliArgs, fenrir_manager=None, environment=environment.environment
    ):
        self.cliArgs = cliArgs

        settings_root = self.get_resource_settings_root()
        if not settings_root:
            return None

        self.save_settings_path = self.get_default_save_settings_file()
        settings_file = self.resolve_settings_file(cliArgs.setting)
        if not settings_file:
            return None

        environment["runtime"]["SettingsManager"] = self
        environment["runtime"]["SettingsManager"].initialize(environment)

        valid_config = environment["runtime"]["SettingsManager"].load_settings(
            settings_file
        )
        if not valid_config:
            return None
        if cliArgs.options != "":
            self.parse_setting_args(cliArgs.options)
        if cliArgs.debug:
            self.set_setting("general", "debug_level", 3)
        if cliArgs.print:
            self.set_setting("general", "debug_level", 3)
            self.set_setting("general", "debug_mode", "PRINT")
        if cliArgs.x11:
            self.set_setting("screen", "driver", "ptyDriver")
            self.set_setting("keyboard", "driver", "x11Driver")
        if cliArgs.x11_window_id:
            self.set_setting(
                "keyboard", "x11_window_id", cliArgs.x11_window_id
            )

        self.set_fenrir_keys(self.get_setting("general", "fenrir_keys"))
        self.set_script_keys(self.get_setting("general", "script_keys"))

        environment["runtime"]["DebugManager"] = debugManager.DebugManager(
            self.env["runtime"]["SettingsManager"].get_setting(
                "general", "debug_file"
            )
        )
        environment["runtime"]["DebugManager"].initialize(environment)

        if cliArgs.force_all_screens:
            environment["runtime"]["force_all_screens"] = True

        if cliArgs.ignore_screen:
            current_ignore_screen = self.get_setting("screen", "ignore_screen")
            if current_ignore_screen:
                ignore_screens = (
                    current_ignore_screen.split(",") + cliArgs.ignore_screen
                )
            else:
                ignore_screens = cliArgs.ignore_screen
            self.set_setting(
                "screen", "ignore_screen", ",".join(ignore_screens)
            )

        sound_theme_path = self.resolve_sound_theme_path(
            self.get_setting("sound", "theme")
        )
        if sound_theme_path:
            self.set_setting("sound", "theme", sound_theme_path)
            environment["runtime"]["SettingsManager"].load_sound_icons(
                self.get_setting("sound", "theme"), environment
            )

        environment["runtime"][
            "PunctuationManager"
        ] = punctuationManager.PunctuationManager()
        environment["runtime"]["PunctuationManager"].initialize(environment)

        environment["runtime"]["TextManager"] = textManager.TextManager()
        environment["runtime"]["TextManager"].initialize(environment)

        if not os.path.exists(
            self.get_setting("general", "punctuation_profile")
        ):
            if os.path.exists(
                settings_root
                + "punctuation/"
                + self.get_setting("general", "punctuation_profile")
            ):
                self.set_setting(
                    "general",
                    "punctuation_profile",
                    settings_root
                    + "punctuation/"
                    + self.get_setting("general", "punctuation_profile"),
                )
                environment["runtime"]["PunctuationManager"].load_dicts(
                    self.get_setting("general", "punctuation_profile")
                )
            if os.path.exists(
                settings_root
                + "punctuation/"
                + self.get_setting("general", "punctuation_profile")
                + ".conf"
            ):
                self.set_setting(
                    "general",
                    "punctuation_profile",
                    settings_root
                    + "punctuation/"
                    + self.get_setting("general", "punctuation_profile")
                    + ".conf",
                )
                environment["runtime"]["PunctuationManager"].load_dicts(
                    self.get_setting("general", "punctuation_profile")
                )
        else:
            environment["runtime"]["PunctuationManager"].load_dicts(
                self.get_setting("general", "punctuation_profile")
            )

        if fenrir_manager:
            environment["runtime"]["FenrirManager"] = fenrir_manager

        environment["runtime"]["MemoryManager"] = memoryManager.MemoryManager()
        environment["runtime"]["MemoryManager"].initialize(environment)

        environment["runtime"][
            "AttributeManager"
        ] = attributeManager.AttributeManager()
        environment["runtime"]["AttributeManager"].initialize(environment)

        environment["runtime"]["EventManager"] = eventManager.EventManager()
        environment["runtime"]["EventManager"].initialize(environment)

        environment["runtime"][
            "ProcessManager"
        ] = processManager.ProcessManager()
        environment["runtime"]["ProcessManager"].initialize(environment)

        environment["runtime"]["OutputManager"] = outputManager.OutputManager()
        environment["runtime"]["OutputManager"].initialize(environment)

        environment["runtime"]["InputManager"] = inputManager.InputManager()
        environment["runtime"]["InputManager"].initialize(environment)
        self.load_keyboard_layout(environment)

        environment["runtime"]["ScreenManager"] = screenManager.ScreenManager()
        environment["runtime"]["ScreenManager"].initialize(environment)

        environment["runtime"][
            "CommandManager"
        ] = commandManager.CommandManager()
        environment["runtime"]["CommandManager"].initialize(environment)

        environment["runtime"]["HelpManager"] = helpManager.HelpManager()
        environment["runtime"]["HelpManager"].initialize(environment)

        environment["runtime"]["RemoteManager"] = remoteManager.RemoteManager()
        environment["runtime"]["RemoteManager"].initialize(environment)

        environment["runtime"][
            "DiffReviewManager"
        ] = diffReviewManager.DiffReviewManager()
        environment["runtime"]["DiffReviewManager"].initialize(environment)

        environment["runtime"]["CursorManager"] = cursorManager.CursorManager()
        environment["runtime"]["CursorManager"].initialize(environment)
        environment["runtime"][
            "ApplicationManager"
        ] = applicationManager.ApplicationManager()
        environment["runtime"]["ApplicationManager"].initialize(environment)
        environment["runtime"]["TextManager"] = textManager.TextManager()
        environment["runtime"]["TextManager"].initialize(environment)
        environment["runtime"]["TableManager"] = tableManager.TableManager()
        environment["runtime"]["TableManager"].initialize(environment)
        environment["runtime"][
            "BarrierManager"
        ] = barrierManager.BarrierManager()
        environment["runtime"]["BarrierManager"].initialize(environment)
        environment["runtime"]["SayAllManager"] = sayAllManager.SayAllManager()
        environment["runtime"]["SayAllManager"].initialize(environment)
        environment["runtime"]["VmenuManager"] = vmenuManager.VmenuManager()
        environment["runtime"]["VmenuManager"].initialize(environment)
        environment["runtime"][
            "QuickMenuManager"
        ] = quickMenuManager.QuickMenuManager()
        environment["runtime"]["QuickMenuManager"].initialize(environment)
        
        environment["runtime"][
            "ReadAllManager"
        ] = readAllManager.ReadAllManager()
        environment["runtime"]["ReadAllManager"].initialize(environment)

        # only possible after having input and screen managers with clean
        # buffer
        environment["runtime"]["InputManager"].write_event_buffer()
        environment["runtime"]["InputManager"].handle_device_grab(force=True)

        environment["runtime"]["DebugManager"].write_debug_out(
            r"/-------environment-------/",
            debug.DebugLevel.INFO,
            on_any_level=True,
        )
        environment["runtime"]["DebugManager"].write_debug_out(
            str(environment), debug.DebugLevel.INFO, on_any_level=True
        )
        environment["runtime"]["DebugManager"].write_debug_out(
            r"/-------settings.conf-------/",
            debug.DebugLevel.INFO,
            on_any_level=True,
        )
        environment["runtime"]["DebugManager"].write_debug_out(
            str(environment["settings"]._sections),
            debug.DebugLevel.INFO,
            on_any_level=True,
        )
        environment["runtime"]["DebugManager"].write_debug_out(
            r"/-------self.settingArgDict-------/",
            debug.DebugLevel.INFO,
            on_any_level=True,
        )
        environment["runtime"]["DebugManager"].write_debug_out(
            str(self.settingArgDict), debug.DebugLevel.INFO, on_any_level=True
        )
        self.bindingsBackup = environment["bindings"].copy()

        return environment



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import os
import re

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _


class DiffReviewManager:
    def __init__(self):
        self.env = None
        self.active = False
        self.loaded_path = ""
        self.lines = []
        self.line_meta = []
        self.file_sections = []
        self.hunk_sections = []
        self.file_starts = []
        self.hunk_starts = []
        self.added_lines = []
        self.removed_lines = []
        self.current_line = -1
        self.bindings_backup = None
        self.raw_bindings_backup = None
        self.total_added = 0
        self.total_removed = 0
        self.current_char = 0

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        if self.active:
            self._restore_bindings()
        self.active = False

    def is_active(self):
        return self.active

    def has_loaded_diff(self):
        return len(self.lines) > 0

    def get_loaded_path(self):
        return self.loaded_path

    def toggle_mode(self):
        if self.active:
            self.disable_mode()
            self.present_mode_message(
                _("Diff review mode disabled."),
            )
            return False

        clipboard_path = self._get_clipboard_path_or_none()
        if (
            clipboard_path
            and (
                (not self.has_loaded_diff())
                or (self.loaded_path != clipboard_path)
            )
        ):
            if not self.load_file(clipboard_path):
                return False
        elif not self.has_loaded_diff():
            self.present_error(
                _("Clipboard does not contain full path to diff file.")
            )
            return False

        enabled = self.enable_mode()
        if enabled:
            self.present_mode_message(
                _("Diff review mode enabled."),
            )
        return enabled

    def enable_mode(self):
        if self.active:
            return True
        if not self.has_loaded_diff():
            return False
        self._install_diff_bindings()
        self.active = True
        return True

    def disable_mode(self):
        if not self.active:
            return
        self._restore_bindings()
        self.active = False

    def load_from_clipboard(self):
        clipboard_path = self._get_clipboard_path_or_none()
        if not clipboard_path:
            self.present_error(
                _("Clipboard does not contain full path to diff file.")
            )
            return False

        return self.load_file(clipboard_path)

    def reload_file(self):
        if not self.loaded_path:
            self.present_error(_("No diff file loaded."))
            return False
        return self.load_file(self.loaded_path)

    def load_file(self, file_path):
        normalized_path = os.path.abspath(os.path.expanduser(file_path))
        if not os.path.exists(normalized_path):
            self.present_error(
                _("Diff file not found: ") + normalized_path
            )
            return False
        if os.path.isdir(normalized_path):
            self.present_error(
                _("Diff path is a directory: ") + normalized_path
            )
            return False
        if not os.access(normalized_path, os.R_OK):
            self.present_error(
                _("Cannot read diff file: ") + normalized_path
            )
            return False

        try:
            with open(normalized_path, "r", encoding="utf-8", errors="replace") as diff_file:
                file_lines = diff_file.read().splitlines()
        except Exception as load_error:
            self._write_debug(
                "DiffReviewManager load_file: " + str(load_error),
                debug.DebugLevel.ERROR,
            )
            self.present_error(
                _("Failed to load diff file: ") + normalized_path
            )
            return False

        self.present_mode_message(
            _("Loading ") + normalized_path + ".",
        )
        self._parse_lines(file_lines, normalized_path)
        self.present_mode_message(
            _("Diff loaded. ")
            + str(len(self.file_sections))
            + _(" files changed. ")
            + str(self.total_added)
            + _(" additions. ")
            + str(self.total_removed)
            + _(" removals."),
        )
        return True

    def _parse_lines(self, file_lines, loaded_path):
        self.lines = file_lines
        self.loaded_path = loaded_path
        self.line_meta = []
        self.file_sections = []
        self.hunk_sections = []
        self.file_starts = []
        self.hunk_starts = []
        self.added_lines = []
        self.removed_lines = []
        self.current_line = -1
        self.total_added = 0
        self.total_removed = 0
        self.current_char = 0

        has_git_header = any(
            line.startswith("diff --git ") for line in self.lines
        )

        current_file_index = -1
        current_hunk_index = -1
        for line_index, line in enumerate(self.lines):
            role = self._classify_line(line)

            if self._starts_new_file_section(
                line=line,
                has_git_header=has_git_header,
            ):
                current_file_index = self._add_file_section(line_index)
                current_hunk_index = -1

            if (
                current_file_index == -1
                and role
                in {
                    "hunk_header",
                    "added_line",
                    "removed_line",
                    "context_line",
                    "note_line",
                }
            ):
                current_file_index = self._add_file_section(line_index)

            if role == "hunk_header":
                if current_file_index == -1:
                    current_file_index = self._add_file_section(line_index)
                current_hunk_index = self._add_hunk_section(
                    line_index=line_index,
                    file_index=current_file_index,
                )

            if role in {"added_line", "removed_line"} and current_hunk_index == -1:
                if current_file_index == -1:
                    current_file_index = self._add_file_section(line_index)
                current_hunk_index = self._add_hunk_section(
                    line_index=line_index,
                    file_index=current_file_index,
                )

            if role == "added_line":
                self.added_lines.append(line_index)
                self.total_added += 1
                if current_file_index != -1:
                    self.file_sections[current_file_index]["added_count"] += 1
                if current_hunk_index != -1:
                    self.hunk_sections[current_hunk_index]["added_count"] += 1
            elif role == "removed_line":
                self.removed_lines.append(line_index)
                self.total_removed += 1
                if current_file_index != -1:
                    self.file_sections[current_file_index]["removed_count"] += 1
                if current_hunk_index != -1:
                    self.hunk_sections[current_hunk_index]["removed_count"] += 1

            self.line_meta.append(
                {
                    "role": role,
                    "file_index": current_file_index,
                    "hunk_index": current_hunk_index,
                }
            )

        for file_index, file_section in enumerate(self.file_sections):
            file_hunks = file_section["hunk_indices"]
            for local_position, hunk_index in enumerate(file_hunks, start=1):
                self.hunk_sections[hunk_index]["position_in_file"] = local_position
                self.hunk_sections[hunk_index]["hunks_in_file"] = len(file_hunks)

    def _classify_line(self, line):
        if line.startswith("\\ No newline at end of file"):
            return "note_line"
        if self._is_file_header(line):
            return "file_header"
        if line.startswith("@@"):
            return "hunk_header"
        if line.startswith("+") and not line.startswith("+++ "):
            return "added_line"
        if line.startswith("-") and not line.startswith("--- "):
            return "removed_line"
        return "context_line"

    def _get_clipboard_path_or_none(self):
        if self.env["runtime"]["MemoryManager"].is_index_list_empty(
            "clipboardHistory"
        ):
            return None

        clipboard_value = self.env["runtime"]["MemoryManager"].get_index_list_element(
            "clipboardHistory"
        )
        if not isinstance(clipboard_value, str):
            return None

        clipboard_path = clipboard_value.strip().strip("\"'")
        clipboard_path = os.path.expanduser(clipboard_path)
        if not os.path.isabs(clipboard_path):
            return None
        return os.path.abspath(clipboard_path)

    def _is_file_header(self, line):
        file_header_prefixes = (
            "diff --git ",
            "--- ",
            "+++ ",
            "index ",
            "old mode ",
            "new mode ",
            "deleted file mode ",
            "new file mode ",
            "rename from ",
            "rename to ",
            "similarity index ",
            "dissimilarity index ",
            "Binary files ",
        )
        return line.startswith(file_header_prefixes)

    def _starts_new_file_section(self, line, has_git_header):
        if has_git_header:
            return line.startswith("diff --git ")
        return line.startswith("--- ")

    def _add_file_section(self, line_index):
        file_section = {
            "start_line": line_index,
            "added_count": 0,
            "removed_count": 0,
            "hunk_indices": [],
        }
        self.file_sections.append(file_section)
        self.file_starts.append(line_index)
        return len(self.file_sections) - 1

    def _add_hunk_section(self, line_index, file_index):
        hunk_section = {
            "start_line": line_index,
            "file_index": file_index,
            "added_count": 0,
            "removed_count": 0,
            "position_in_file": 0,
            "hunks_in_file": 0,
        }
        self.hunk_sections.append(hunk_section)
        self.hunk_starts.append(line_index)
        hunk_index = len(self.hunk_sections) - 1
        if file_index != -1:
            self.file_sections[file_index]["hunk_indices"].append(hunk_index)
        return hunk_index

    def next_file_section(self):
        self._jump_to_anchor(
            anchors=self.file_starts,
            previous=False,
            empty_message=_("No file sections in loaded diff."),
            edge_message=_("No further file sections."),
            announce_role="file",
        )

    def prev_file_section(self):
        self._jump_to_anchor(
            anchors=self.file_starts,
            previous=True,
            empty_message=_("No file sections in loaded diff."),
            edge_message=_("No previous file sections."),
            announce_role="file",
        )

    def next_hunk(self):
        self._jump_to_anchor(
            anchors=self.hunk_starts,
            previous=False,
            empty_message=_("No hunks in loaded diff."),
            edge_message=_("No further hunks."),
            announce_role="hunk",
        )

    def prev_hunk(self):
        self._jump_to_anchor(
            anchors=self.hunk_starts,
            previous=True,
            empty_message=_("No hunks in loaded diff."),
            edge_message=_("No previous hunks."),
            announce_role="hunk",
        )

    def next_added(self):
        self._jump_to_anchor(
            anchors=self.added_lines,
            previous=False,
            empty_message=_("No added lines in loaded diff."),
            edge_message=_("No further added lines."),
            announce_role="added",
        )

    def prev_added(self):
        self._jump_to_anchor(
            anchors=self.added_lines,
            previous=True,
            empty_message=_("No added lines in loaded diff."),
            edge_message=_("No previous added lines."),
            announce_role="added",
        )

    def next_removed(self):
        self._jump_to_anchor(
            anchors=self.removed_lines,
            previous=False,
            empty_message=_("No removed lines in loaded diff."),
            edge_message=_("No further removed lines."),
            announce_role="removed",
        )

    def prev_removed(self):
        self._jump_to_anchor(
            anchors=self.removed_lines,
            previous=True,
            empty_message=_("No removed lines in loaded diff."),
            edge_message=_("No previous removed lines."),
            announce_role="removed",
        )

    def present_summary(self):
        if not self.has_loaded_diff():
            self.present_mode_message(_("No diff file loaded."))
            return

        summary = (
            _("Diff summary. Files changed: ")
            + str(len(self.file_sections))
            + _(". Additions: ")
            + str(self.total_added)
            + _(". Removals: ")
            + str(self.total_removed)
            + "."
        )

        if self.current_line >= 0 and self.current_line < len(self.line_meta):
            location = self._build_location_text(self.current_line)
            if location:
                summary += " " + location

        self.present_mode_message(summary)

    def present_key_help(self):
        help_text = _(
            "Diff keys: h next hunk, shift h previous hunk, a next addition, "
            "shift a previous addition, d next removal, shift d previous removal, "
            "f next file section, shift f previous file section, up previous line, "
            "down next line, left previous character, right next character, "
            "control left previous word, control right next word, s summary, "
            "escape exit diff mode."
        )
        self.present_mode_message(help_text)

    def next_line(self):
        if not self.has_loaded_diff():
            self.present_mode_message(_("No diff file loaded."))
            return
        if self.current_line < 0:
            self.current_line = 0
            self.current_char = 0
            self._announce_current_line()
            return
        if self.current_line >= len(self.lines) - 1:
            self.present_mode_message(_("End of diff."))
            return
        self.current_line += 1
        self.current_char = 0
        self._announce_current_line()

    def prev_line(self):
        if not self.has_loaded_diff():
            self.present_mode_message(_("No diff file loaded."))
            return
        if self.current_line < 0:
            self.current_line = 0
            self.current_char = 0
            self._announce_current_line()
            return
        if self.current_line <= 0:
            self.present_mode_message(_("Start of diff."))
            return
        self.current_line -= 1
        self.current_char = 0
        self._announce_current_line()

    def next_char(self):
        if not self._ensure_current_line():
            return
        line_text = self.lines[self.current_line]
        if len(line_text) == 0:
            self.present_mode_message(_("blank"))
            return
        if self.current_char >= len(line_text) - 1:
            self.present_mode_message(_("End of line."))
            return
        self.current_char += 1
        self._announce_current_char()

    def prev_char(self):
        if not self._ensure_current_line():
            return
        line_text = self.lines[self.current_line]
        if len(line_text) == 0:
            self.present_mode_message(_("blank"))
            return
        if self.current_char <= 0:
            self.present_mode_message(_("Start of line."))
            return
        self.current_char -= 1
        self._announce_current_char()

    def next_word(self):
        if not self._ensure_current_line():
            return
        line_text = self.lines[self.current_line]
        word_spans = self._get_word_spans(line_text)
        if len(word_spans) == 0:
            self.present_mode_message(_("No words on this line."))
            return

        next_word_start = None
        for start, end in word_spans:
            if start > self.current_char:
                next_word_start = start
                break
            if start <= self.current_char < end:
                continue

        if next_word_start is None:
            self.present_mode_message(_("No next word on this line."))
            return

        self.current_char = next_word_start
        self._announce_word_at_current_char()

    def prev_word(self):
        if not self._ensure_current_line():
            return
        line_text = self.lines[self.current_line]
        word_spans = self._get_word_spans(line_text)
        if len(word_spans) == 0:
            self.present_mode_message(_("No words on this line."))
            return

        prev_word_start = None
        for start, end in word_spans:
            if end - 1 < self.current_char:
                prev_word_start = start
            elif start <= self.current_char < end:
                break
            else:
                break

        if prev_word_start is None:
            self.present_mode_message(_("No previous word on this line."))
            return

        self.current_char = prev_word_start
        self._announce_word_at_current_char()

    def _jump_to_anchor(
        self,
        anchors,
        previous,
        empty_message,
        edge_message,
        announce_role,
    ):
        if not self.has_loaded_diff():
            self.present_mode_message(_("No diff file loaded."))
            return

        if len(anchors) == 0:
            self.present_mode_message(empty_message)
            return

        new_line = self._find_next_anchor(anchors, previous)
        if new_line is None:
            self.present_mode_message(edge_message)
            return

        self.current_line = new_line
        self.current_char = 0
        self._announce_line(announce_role, new_line)

    def _find_next_anchor(self, anchors, previous):
        if self.current_line < 0:
            if previous:
                return anchors[-1]
            return anchors[0]

        if previous:
            for anchor_line in reversed(anchors):
                if anchor_line < self.current_line:
                    return anchor_line
            return None

        for anchor_line in anchors:
            if anchor_line > self.current_line:
                return anchor_line
        return None

    def _announce_line(self, announce_role, line_index):
        line_text = self.lines[line_index] if line_index < len(self.lines) else ""
        line_meta = self.line_meta[line_index]
        role = line_meta["role"]
        verbosity = self._get_diff_verbosity()

        if announce_role == "file":
            self._announce_file_section(line_index, line_meta)
            return
        if announce_role == "hunk":
            self._announce_hunk_section(line_index, line_meta, line_text)
            return

        if role == "added_line":
            base_message = _("Added line.")
        elif role == "removed_line":
            base_message = _("Removed line.")
        else:
            base_message = _("Diff line.")

        message = base_message + " " + self._build_location_text(line_index)
        if verbosity == "verbose":
            content = self._line_content_for_speech(line_text, role)
            message = message + " " + content
        self.present_mode_message(message)

    def _announce_current_line(self):
        if self.current_line < 0 or self.current_line >= len(self.lines):
            self.present_mode_message(_("No diff file loaded."))
            return
        line_text = self.lines[self.current_line]
        role = self.line_meta[self.current_line]["role"]

        if role == "context_line":
            formatted_classic_diff = self._format_classic_diff_command(line_text)
            if formatted_classic_diff:
                self.present_mode_message(formatted_classic_diff)
                return

        role_prefix = {
            "added_line": _("Added"),
            "removed_line": _("Removed"),
            "hunk_header": _("Hunk header"),
            "file_header": _("File header"),
            "note_line": _("Note"),
            "context_line": "",
        }.get(role, _("Line"))
        if line_text.strip() == "":
            if role_prefix:
                self.present_mode_message(role_prefix + ": " + _("blank"))
            else:
                self.present_mode_message(_("blank"))
            return
        if role_prefix:
            line_text = self._normalize_diff_line_content(line_text, role)
            if line_text.strip() == "":
                self.present_mode_message(role_prefix + ": " + _("blank"))
                return
            if role in {"added_line", "removed_line"} and line_text in {
                _("added"),
                _("removed"),
            }:
                self.present_mode_message(line_text)
                return
            self.present_mode_message(role_prefix + ": " + line_text)
        else:
            self.present_mode_message(line_text)

    def _format_classic_diff_command(self, line_text):
        match = re.match(r"^(\d+)(?:,(\d+))?([acd])(\d+)(?:,(\d+))?$", line_text.strip())
        if not match:
            return None

        old_start = int(match.group(1))
        old_end = int(match.group(2) or match.group(1))
        op = match.group(3)
        new_start = int(match.group(4))
        new_end = int(match.group(5) or match.group(4))

        if op == "c":
            return self._format_line_span(old_start, old_end) + " " + _("changed")
        if op == "d":
            return self._format_line_span(old_start, old_end) + " " + _("deleted")
        if op == "a":
            return self._format_line_span(new_start, new_end) + " " + _("added")
        return None

    def _format_line_span(self, start_line, end_line):
        if start_line == end_line:
            return _("line ") + str(start_line)
        return (
            _("lines ")
            + str(start_line)
            + _(" through ")
            + str(end_line)
        )

    def _announce_current_char(self):
        if self.current_line < 0 or self.current_line >= len(self.lines):
            self.present_mode_message(_("No diff file loaded."))
            return
        line_text = self.lines[self.current_line]
        if len(line_text) == 0:
            self.present_mode_message(_("blank"))
            return
        if self.current_char < 0:
            self.current_char = 0
        if self.current_char >= len(line_text):
            self.current_char = len(line_text) - 1
        char_value = line_text[self.current_char]
        if char_value == " ":
            self.present_mode_message(_("space"))
            return
        self.env["runtime"]["OutputManager"].present_text(
            char_value,
            interrupt=True,
            ignore_punctuation=True,
            announce_capital=True,
        )

    def _announce_word_at_current_char(self):
        if self.current_line < 0 or self.current_line >= len(self.lines):
            self.present_mode_message(_("No diff file loaded."))
            return
        line_text = self.lines[self.current_line]
        for start, end in self._get_word_spans(line_text):
            if start <= self.current_char < end:
                self.present_mode_message(line_text[start:end])
                return
        self.present_mode_message(_("No word at current position."))

    def _ensure_current_line(self):
        if not self.has_loaded_diff():
            self.present_mode_message(_("No diff file loaded."))
            return False
        if self.current_line < 0:
            self.current_line = 0
            self.current_char = 0
        if self.current_line >= len(self.lines):
            self.current_line = len(self.lines) - 1
            self.current_char = 0
        return True

    def _get_word_spans(self, line_text):
        return [match.span() for match in re.finditer(r"\S+", line_text)]

    def _announce_file_section(self, line_index, line_meta):
        file_index = line_meta["file_index"]
        if file_index == -1 or file_index >= len(self.file_sections):
            self.present_mode_message(_("File section."))
            return
        file_section = self.file_sections[file_index]
        message = (
            _("File section ")
            + str(file_index + 1)
            + _(" of ")
            + str(len(self.file_sections))
            + _(". Plus ")
            + str(file_section["added_count"])
            + _(", minus ")
            + str(file_section["removed_count"])
            + "."
        )
        self.present_mode_message(message)

    def _announce_hunk_section(self, line_index, line_meta, line_text):
        hunk_index = line_meta["hunk_index"]
        if hunk_index == -1 or hunk_index >= len(self.hunk_sections):
            self.present_mode_message(_("Hunk section."))
            return

        hunk_section = self.hunk_sections[hunk_index]
        file_index = hunk_section["file_index"]
        file_position = file_index + 1 if file_index >= 0 else 0
        hunk_position = hunk_section["position_in_file"]
        hunk_total = hunk_section["hunks_in_file"]

        message = (
            _("Hunk ")
            + str(hunk_position)
            + _(" of ")
            + str(hunk_total)
            + _(" in file ")
            + str(file_position)
            + _(" of ")
            + str(len(self.file_sections))
            + _(". Plus ")
            + str(hunk_section["added_count"])
            + _(", minus ")
            + str(hunk_section["removed_count"])
            + "."
        )

        if self._get_diff_verbosity() == "verbose":
            message += " " + self._line_content_for_speech(line_text, "hunk_header")

        self.present_mode_message(message)

    def _line_content_for_speech(self, line_text, role):
        line_text = self._normalize_diff_line_content(line_text, role)
        if line_text.strip() == "":
            return _("blank")
        return line_text

    def _normalize_diff_line_content(self, line_text, role):
        if role == "added_line":
            stripped_line = line_text.lstrip("+")
            if stripped_line.strip() == "" and line_text.startswith("+"):
                return _("added")
            return stripped_line
        if role == "removed_line":
            stripped_line = line_text.lstrip("-")
            if stripped_line.strip() == "" and line_text.startswith("-"):
                return _("removed")
            return stripped_line
        return line_text

    def _build_location_text(self, line_index):
        if line_index < 0 or line_index >= len(self.line_meta):
            return ""
        line_meta = self.line_meta[line_index]
        file_index = line_meta["file_index"]
        hunk_index = line_meta["hunk_index"]
        location_parts = []

        if file_index != -1 and file_index < len(self.file_sections):
            location_parts.append(
                _("File ")
                + str(file_index + 1)
                + _(" of ")
                + str(len(self.file_sections))
                + ".")

        if hunk_index != -1 and hunk_index < len(self.hunk_sections):
            hunk_info = self.hunk_sections[hunk_index]
            location_parts.append(
                _("Hunk ")
                + str(hunk_info["position_in_file"])
                + _(" of ")
                + str(hunk_info["hunks_in_file"])
                + "."
            )

        return " ".join(location_parts)

    def _get_diff_verbosity(self):
        verbosity = self.env["runtime"]["SettingsManager"].get_setting(
            "general", "diff_verbosity"
        ).lower()
        if verbosity not in {"compact", "verbose"}:
            return "compact"
        return verbosity

    def present_error(self, message):
        self.present_mode_message(message)

    def present_mode_message(self, message):
        self.env["runtime"]["OutputManager"].present_text(
            message,
            interrupt=True,
        )

    def _install_diff_bindings(self):
        self.bindings_backup = self.env["bindings"].copy()
        self.raw_bindings_backup = self.env["rawBindings"].copy()

        diff_bindings = {
            str([1, ["KEY_H"]]): "DIFF_NEXT_HUNK",
            str([1, sorted(["KEY_SHIFT", "KEY_H"])]): "DIFF_PREV_HUNK",
            str([1, ["KEY_A"]]): "DIFF_NEXT_ADDED",
            str([1, sorted(["KEY_SHIFT", "KEY_A"])]): "DIFF_PREV_ADDED",
            str([1, ["KEY_D"]]): "DIFF_NEXT_REMOVED",
            str([1, sorted(["KEY_SHIFT", "KEY_D"])]): "DIFF_PREV_REMOVED",
            str([1, ["KEY_F"]]): "DIFF_NEXT_FILE_SECTION",
            str([1, sorted(["KEY_SHIFT", "KEY_F"])]): "DIFF_PREV_FILE_SECTION",
            str([1, ["KEY_S"]]): "DIFF_SUMMARY",
            str([1, ["KEY_F1"]]): "DIFF_KEY_HELP",
            str([1, ["KEY_UP"]]): "DIFF_PREV_LINE",
            str([1, ["KEY_DOWN"]]): "DIFF_NEXT_LINE",
            str([1, ["KEY_LEFT"]]): "DIFF_PREV_CHAR",
            str([1, ["KEY_RIGHT"]]): "DIFF_NEXT_CHAR",
            str([1, sorted(["KEY_CTRL", "KEY_LEFT"])]): "DIFF_PREV_WORD",
            str([1, sorted(["KEY_CTRL", "KEY_RIGHT"])]): "DIFF_NEXT_WORD",
            str([1, ["KEY_ESC"]]): "EXIT_DIFF_MODE",
            str([1, sorted(["KEY_FENRIR", "KEY_CTRL", "KEY_D"])]): "TOGGLE_DIFF_MODE",
        }

        diff_raw_bindings = {
            shortcut: [1, keys]
            for shortcut, keys in [
                (str([1, ["KEY_H"]]), ["KEY_H"]),
                (str([1, sorted(["KEY_SHIFT", "KEY_H"])]), sorted(["KEY_SHIFT", "KEY_H"])),
                (str([1, ["KEY_A"]]), ["KEY_A"]),
                (str([1, sorted(["KEY_SHIFT", "KEY_A"])]), sorted(["KEY_SHIFT", "KEY_A"])),
                (str([1, ["KEY_D"]]), ["KEY_D"]),
                (str([1, sorted(["KEY_SHIFT", "KEY_D"])]), sorted(["KEY_SHIFT", "KEY_D"])),
                (str([1, ["KEY_F"]]), ["KEY_F"]),
                (str([1, sorted(["KEY_SHIFT", "KEY_F"])]), sorted(["KEY_SHIFT", "KEY_F"])),
                (str([1, ["KEY_S"]]), ["KEY_S"]),
                (str([1, ["KEY_F1"]]), ["KEY_F1"]),
                (str([1, ["KEY_UP"]]), ["KEY_UP"]),
                (str([1, ["KEY_DOWN"]]), ["KEY_DOWN"]),
                (str([1, ["KEY_LEFT"]]), ["KEY_LEFT"]),
                (str([1, ["KEY_RIGHT"]]), ["KEY_RIGHT"]),
                (
                    str([1, sorted(["KEY_CTRL", "KEY_LEFT"])]),
                    sorted(["KEY_CTRL", "KEY_LEFT"]),
                ),
                (
                    str([1, sorted(["KEY_CTRL", "KEY_RIGHT"])]),
                    sorted(["KEY_CTRL", "KEY_RIGHT"]),
                ),
                (str([1, ["KEY_ESC"]]), ["KEY_ESC"]),
                (
                    str([1, sorted(["KEY_FENRIR", "KEY_CTRL", "KEY_D"])]),
                    sorted(["KEY_FENRIR", "KEY_CTRL", "KEY_D"]),
                ),
            ]
        }

        self.env["bindings"] = diff_bindings
        self.env["rawBindings"] = diff_raw_bindings

    def _restore_bindings(self):
        if self.bindings_backup is not None:
            self.env["bindings"] = self.bindings_backup
        if self.raw_bindings_backup is not None:
            self.env["rawBindings"] = self.raw_bindings_backup
        self.bindings_backup = None
        self.raw_bindings_backup = None

    def _write_debug(self, message, level):
        self.env["runtime"]["DebugManager"].write_debug_out(
            message, level
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug

runtime_data = {
    "SpeechDriver": None,
    "ScreenDriver": None,
    "SoundDriver": None,
    "InputDriver": None,
    "RemoteDriver": None,
    "InputManager": None,
    "CommandManager": None,
    "ScreenManager": None,
    "OutputManager": None,
    "DebugManager": None,
    "SettingsManager": None,
    "FenrirManager": None,
    "EventManager": None,
    "ProcessManager": None,
    "DiffReviewManager": None,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug


class SayAllManager:
    def __init__(self):
        self.isActive = False
        self.isActiveLock = None

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def set_is_active(self, isActive):
        pass

    def start(self):
        pass

    def is_say_all_active(self):
        pass

    def say_all_worker(self):
        pass

    def stop(self):
        pass

    def finish(self):
        pass

    def goto_next_page(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.core.tabCompletionManager import TabCompletionManager


class command:
    def __init__(self):
        self.manager = TabCompletionManager()

    def initialize(self, environment):
        self.env = environment
        self.manager.initialize(environment)

    def shutdown(self):
        pass

    def get_description(self):
        return _("Announces tab completions on cursor movement")

    def run(self):
        text = self.manager.process_update()
        if not text:
            return

        do_interrupt = True
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            do_interrupt = False

        self.env["runtime"]["OutputManager"].present_text(
            text,
            interrupt=do_interrupt,
            announce_capital=True,
            flush=False,
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import _thread
import importlib
import os

import pyperclip

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment, script_path=""):
        self.env = environment
        self.script_path = script_path

    def shutdown(self):
        pass

    def get_description(self):
        return _("imports the graphical clipboard to Fenrir's clipboard")

    def run(self):
        _thread.start_new_thread(self._thread_run, ())

    def _thread_run(self):
        try:
            # Remember original display environment variable if it exists
            original_display = os.environ.get("DISPLAY", "")
            clipboard_content = None

            # Try different display options
            for i in range(10):
                display = f":{i}"
                try:
                    # Set display environment variable
                    os.environ["DISPLAY"] = display
                    # Attempt to get clipboard content
                    # Weird workaround for some distros
                    importlib.reload(pyperclip)
                    clipboard_content = pyperclip.paste()
                    # If we get here without exception, we found a working
                    # display
                    if clipboard_content:
                        break
                except Exception:
                    # Failed for this display, try next one
                    continue

            # Restore original display setting
            if original_display:
                os.environ["DISPLAY"] = original_display
            else:
                os.environ.pop("DISPLAY", None)

            # Process the clipboard content if we found any
            if clipboard_content and isinstance(clipboard_content, str):
                self.env["runtime"]["MemoryManager"].add_value_to_first_index(
                    "clipboardHistory", clipboard_content
                )
                self.env["runtime"]["OutputManager"].present_text(
                    "Import to Clipboard",
                    sound_icon="CopyToClipboard",
                    interrupt=True,
                )
                self.env["runtime"]["OutputManager"].present_text(
                    clipboard_content, sound_icon="", interrupt=False
                )
            else:
                self.env["runtime"]["OutputManager"].present_text(
                    "No text found in clipboard or no accessible display",
                    interrupt=True,
                )
        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                str(e), sound_icon="", interrupt=False
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import string

from fenrirscreenreader.core import debug


def get_current_word(curr_x, curr_y, curr_text):
    line_break = False
    end_of_screen = False
    if curr_text == "":
        return -1, -1, "", end_of_screen, line_break
    if curr_text.strip(string.whitespace) == "":
        return curr_x, curr_y, "", end_of_screen, line_break
    x = curr_x
    y = curr_y
    curr_word = ""
    wrapped_lines = curr_text.split("\n")
    curr_line = wrapped_lines[y]
    found = False
    while not found:
        if not curr_line[x] in string.whitespace:
            if x == 0:
                found = True
            else:
                if curr_line[x - 1] in string.whitespace:
                    found = True
        if not found:
            if x - 1 < 0:
                if y - 1 < 0:
                    line_break = False
                    end_of_screen = True
                    return curr_x, curr_y, "", end_of_screen, line_break
                else:
                    y -= 1
                    curr_line = wrapped_lines[y]
                    x = len(curr_line) - 1
                    line_break = True
            else:
                x -= 1
    if found:
        curr_word = curr_line[x:]
        for d in string.whitespace:
            delimiter_pos = curr_word.find(d)
            if delimiter_pos != -1:
                curr_word = curr_word[:delimiter_pos]
        return x, y, curr_word, end_of_screen, line_break
    return curr_x, curr_y, "", False, False


def get_prev_word(curr_x, curr_y, curr_text):
    line_break = False
    end_of_screen = False
    if curr_text == "":
        return -1, -1, "", end_of_screen, line_break
    if curr_text.strip(string.whitespace) == "":
        return curr_x, curr_y, "", end_of_screen, line_break
    x, y, curr_word, end_of_screen, line_break_curr_word = get_current_word(
        curr_x, curr_y, curr_text
    )
    if end_of_screen:
        return x, y, curr_word, end_of_screen, line_break
    wrapped_lines = curr_text.split("\n")
    curr_line = wrapped_lines[y]
    if x - 1 < 0:
        if y - 1 < 0:
            line_break = False
            end_of_screen = True
            return curr_x, curr_y, "", end_of_screen, line_break
        else:
            y -= 1
            curr_line = wrapped_lines[y]
            x = len(curr_line) - 1
            line_break = True
    else:
        x -= 1
    line_break_curr_word = line_break or line_break_curr_word
    x, y, curr_word, end_of_screen, line_break = get_current_word(
        x, y, curr_text
    )
    line_break = line_break or line_break_curr_word
    return x, y, curr_word, end_of_screen, line_break


def get_next_word(curr_x, curr_y, curr_text):
    line_break = False
    end_of_screen = False
    if curr_text == "":
        return -1, -1, "", end_of_screen, line_break
    if curr_text.strip(string.whitespace) == "":
        return curr_x, curr_y, "", end_of_screen, line_break
    x = curr_x
    y = curr_y
    curr_word = ""
    wrapped_lines = curr_text.split("\n")
    curr_line = wrapped_lines[y]
    found = False
    while not found:
        if not found:
            if x + 1 > len(curr_line) - 1:
                if y + 1 > len(wrapped_lines) - 1:
                    line_break = False
                    end_of_screen = True
                    return curr_x, curr_y, "", end_of_screen, line_break
                else:
                    y += 1
                    curr_line = wrapped_lines[y]
                    x = 0
                    line_break = True
            else:
                x += 1
        if not curr_line[x] in string.whitespace:
            if x == 0:
                found = True
            else:
                if curr_line[x - 1] in string.whitespace:
                    found = True
    if found:
        curr_word = curr_line[x:]
        for d in string.whitespace:
            delimiter_pos = curr_word.find(d)
            if delimiter_pos != -1:
                curr_word = curr_word[:delimiter_pos]
        return x, y, curr_word, end_of_screen, line_break
    return curr_x, curr_y, "", False, False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return ""

    def run(self):
        if self.env["screen"]["newAttribDelta"] != "":
            return
        if self.env["runtime"]["ScreenManager"].is_screen_change():
            return
        if self.env["runtime"]["CursorManager"].is_cursor_vertical_move():
            return
        if self.env["runtime"]["AttributeManager"].is_attribute_change():
            return
        # hack for pdmenu and maybe other dialog apps that place the cursor at 
        # last cell/row
        # this is not to be identified as history
        if (
            self.env["screen"]["new_cursor"]["x"]
            == self.env["runtime"]["ScreenManager"].get_columns() - 1
        ) and (
            self.env["screen"]["new_cursor"]["y"]
            == self.env["runtime"]["ScreenManager"].get_rows() - 1
        ):
            return
        if not (
            self.env["runtime"]["InputManager"].get_last_deepest_input()
            in [["KEY_UP"], ["KEY_DOWN"]]
        ):
            return

        # Get the current cursor's line from both old and new content
        prev_line = self.env["screen"]["old_content_text"].split("\n")[
            self.env["screen"]["new_cursor"]["y"]
        ]
        curr_line = self.env["screen"]["new_content_text"].split("\n")[
            self.env["screen"]["new_cursor"]["y"]
        ]

        is_blank = curr_line.strip() == ""

        if prev_line == curr_line:
            if self.env["screen"]["new_delta"] != "":
                return

        announce = curr_line
        if not is_blank:
            curr_prompt = curr_line.find("$")
            root_prompt = curr_line.find("#")
            if curr_prompt <= 0:
                if root_prompt > 0:
                    curr_prompt = root_prompt
                else:
                    announce = curr_line
            if curr_prompt > 0:
                remove_digits = str.maketrans("0123456789", "          ")
                if prev_line[:curr_prompt].translate(
                    remove_digits
                ) == curr_line[:curr_prompt].translate(remove_digits):
                    announce = curr_line[curr_prompt + 1 :]
                else:
                    announce = curr_line

        if is_blank:
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), sound_icon="EmptyLine", interrupt=True, flush=False
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                announce, interrupt=True, flush=False
            )

        self.env["commandsIgnore"]["onScreenUpdate"]["CHAR_DELETE_ECHO"] = True
        self.env["commandsIgnore"]["onScreenUpdate"]["CHAR_ECHO"] = True
        self.env["commandsIgnore"]["onScreenUpdate"]["INCOMING_IGNORE"] = True

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("enables or disables speech")

    def run(self):
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "enabled"
        ):
            self.env["runtime"]["OutputManager"].interrupt_output()
            self.env["runtime"]["OutputManager"].present_text(
                _("speech disabled"), sound_icon="SpeechOff", interrupt=True
            )
        self.env["runtime"]["SettingsManager"].set_setting(
            "speech",
            "enabled",
            str(
                not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                    "speech", "enabled"
                )
            ),
        )
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "enabled"
        ):
            self.env["runtime"]["OutputManager"].present_text(
                _("speech enabled"), sound_icon="SpeechOn", interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug
import threading
import time


class ReadAllManager:
    """Manages continuous read-all functionality with speech completion callbacks."""
    
    def __init__(self):
        self.active = False
        self.mode = None  # 'line' or 'page'
        self.stop_requested = False
        self.env = None
        self.items_read = 0
        self._lock = threading.Lock()
    
    def initialize(self, environment):
        self.env = environment
    
    def shutdown(self):
        self.stop_read_all()
    
    def start_read_all(self, mode):
        """Start continuous reading in line or page mode."""
        with self._lock:
            if self.active:
                self.stop_read_all()
            
            self.active = True
            self.mode = mode
            self.stop_requested = False
            self.items_read = 0
            
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"ReadAllManager: Starting read-all mode '{mode}'",
                debug.DebugLevel.INFO
            )
            
            # Send first navigation key
            self._send_navigation_key()
    
    def stop_read_all(self):
        """Stop continuous reading."""
        with self._lock:
            if not self.active:
                return
                
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"ReadAllManager: Stopping read-all after {self.items_read} {self.mode}s",
                debug.DebugLevel.INFO
            )
            
            self.active = False
            self.stop_requested = True
            self.mode = None
            
            # Cancel any pending speech
            try:
                self.env['runtime']['SpeechDriver'].cancel()
            except Exception as e:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"ReadAllManager: Error canceling speech: {e}",
                    debug.DebugLevel.ERROR
                )
    
    def is_active(self):
        """Check if read-all is currently active."""
        return self.active
    
    def get_mode(self):
        """Get current read-all mode ('line' or 'page')."""
        return self.mode
    
    def _send_navigation_key(self):
        """Send appropriate navigation key based on mode."""
        if not self.active or self.stop_requested:
            return
            
        try:
            if self.mode == 'line':
                key_sequence = [[1, 'KEY_DOWN'], [0, 'KEY_DOWN']]
                key_name = 'KEY_DOWN'
            elif self.mode == 'page':
                key_sequence = [[1, 'KEY_PAGEDOWN'], [0, 'KEY_PAGEDOWN']]
                key_name = 'KEY_PAGEDOWN'
            else:
                return
                
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"ReadAllManager: Sending {key_name} ({self.items_read + 1})",
                debug.DebugLevel.INFO
            )
            
            self.env['runtime']['InputManager'].send_keys(key_sequence)
            self.items_read += 1
            
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"ReadAllManager: Error sending key: {e}",
                debug.DebugLevel.ERROR
            )
            self.stop_read_all()
    
    def speech_completed(self):
        """Called when speech completes - continue reading if active."""
        if not self.active or self.stop_requested:
            return
            
        self.env["runtime"]["DebugManager"].write_debug_out(
            f"ReadAllManager: Speech completed, continuing read-all",
            debug.DebugLevel.INFO
        )
        
        # Small delay to prevent overwhelming the system
        def continue_reading():
            time.sleep(0.1)
            if self.active and not self.stop_requested:
                self._send_navigation_key()
        
        # Use daemon thread to avoid blocking
        thread = threading.Thread(target=continue_reading, daemon=True)
        thread.start()



#!/usr/bin/env python3

import importlib.util
import os
import subprocess
import time

from fenrirscreenreader.core import debug


class DynamicVoiceCommand:
    """Dynamic command class for voice selection"""

    def __init__(self, module, voice, env):
        self.module = module
        self.voice = voice
        self.env = env
        self.testMessage = "This is a voice test. The quick brown fox jumps over the lazy dog."

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return f"Select voice: {self.voice}"

    def run(self):
        try:
            self.env["runtime"]["OutputManager"].present_text(
                f"Testing voice {self.voice} from {self.module}. Please wait.",
                interrupt=True,
            )

            # Brief pause before testing to avoid speech overlap
            time.sleep(0.5)

            # Test voice
            testResult, errorMsg = self.test_voice()
            if testResult:
                self.env["runtime"]["OutputManager"].present_text(
                    "Voice test completed successfully. Navigate to Apply Tested Voice to use this voice.",
                    interrupt=False,
                    flush=False,
                )

                # Store for confirmation (use same variables as
                # apply_tested_voice.py)
                self.env["commandBuffer"]["lastTestedModule"] = self.module
                self.env["commandBuffer"]["lastTestedVoice"] = self.voice
                self.env["commandBuffer"]["pendingVoiceModule"] = self.module
                self.env["commandBuffer"]["pendingVoiceVoice"] = self.voice
                self.env["commandBuffer"]["voiceTestCompleted"] = True
            else:
                self.env["runtime"]["OutputManager"].present_text(
                    f"Voice test failed: {errorMsg}",
                    interrupt=False,
                    flush=False,
                )

        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                f"Voice selection error: {str(e)}",
                interrupt=False,
                flush=False,
            )

    def test_voice(self):
        """Test voice with spd-say"""
        try:
            cmd = [
                "spd-say",
                "-C",
                "-w",
                "-o",
                self.module,
                "-y",
                self.voice,
                self.testMessage,
            ]
            result = subprocess.run(
                cmd, timeout=8, capture_output=True, text=True
            )
            if result.returncode == 0:
                return True, "Voice test successful"
            else:
                error_msg = (
                    result.stderr.strip()
                    if result.stderr
                    else f"Command failed with return code {result.returncode}"
                )
                return False, error_msg
        except subprocess.TimeoutExpired:
            return False, "Voice test timed out"
        except Exception as e:
            return False, f"Error running voice test: {str(e)}"

    def set_callback(self, callback):
        pass


class DynamicApplyVoiceCommand:
    """Command to apply the tested voice"""

    def __init__(self, env):
        self.env = env

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "Apply tested voice to Fenrir"

    def run(self):
        try:
            if not self.env["commandBuffer"].get("voiceTestCompleted", False):
                self.env["runtime"]["OutputManager"].present_text(
                    "No voice has been tested yet", interrupt=True
                )
                return

            module = self.env["commandBuffer"]["pendingVoiceModule"]
            voice = self.env["commandBuffer"]["pendingVoiceVoice"]

            self.env["runtime"]["OutputManager"].present_text(
                f"Applying {voice} from {module}", interrupt=True
            )

            # Debug: Show current settings
            settings_manager = self.env["runtime"]["SettingsManager"]
            current_module = settings_manager.get_setting("speech", "module")
            current_voice = settings_manager.get_setting("speech", "voice")
            self.env["runtime"]["OutputManager"].present_text(
                f"Current: {current_voice} from {current_module}",
                interrupt=False,
                flush=False,
            )

            # Apply to runtime settings with fallback
            settings_manager = self.env["runtime"]["SettingsManager"]

            # Store old values for safety
            old_driver = settings_manager.get_setting("speech", "driver")
            old_module = settings_manager.get_setting("speech", "module")
            old_voice = settings_manager.get_setting("speech", "voice")

            try:
                # Apply new settings to runtime only (use set_setting to update
                # settingArgDict)
                settings_manager.set_setting(
                    "speech", "driver", "speechdDriver"
                )
                settings_manager.set_setting("speech", "module", module)
                settings_manager.set_setting("speech", "voice", voice)

                # Apply settings to speech driver directly
                if "SpeechDriver" in self.env["runtime"]:
                    SpeechDriver = self.env["runtime"]["SpeechDriver"]

                    # Get current module to see if we're changing modules
                    current_module = settings_manager.get_setting(
                        "speech", "module"
                    )
                    module_changing = current_module != module

                    # Set module and voice on driver instance first
                    SpeechDriver.set_module(module)
                    SpeechDriver.set_voice(voice)

                    if module_changing:
                        # Module change requires reinitializing the speech
                        # driver
                        self.env["runtime"]["OutputManager"].present_text(
                            f"Switching from {current_module} to {module} module",
                            interrupt=True,
                        )
                        SpeechDriver.shutdown()
                        SpeechDriver.initialize(self.env)
                        # Re-set after initialization
                        SpeechDriver.set_module(module)
                        SpeechDriver.set_voice(voice)
                        self.env["runtime"]["OutputManager"].present_text(
                            "Speech driver reinitialized", interrupt=True
                        )

                    # Debug: verify what was actually set
                    self.env["runtime"]["OutputManager"].present_text(
                        f"Speech driver now has module: {SpeechDriver.module}, voice: {SpeechDriver.voice}",
                        interrupt=True,
                    )

                    # Force application by speaking a test message
                    self.env["runtime"]["OutputManager"].present_text(
                        "Voice applied successfully! You should hear this in the new voice.",
                        interrupt=True,
                    )

                    # Brief pause then more speech to test
                    time.sleep(1)
                    self.env["runtime"]["OutputManager"].present_text(
                        "Use save settings to make permanent", interrupt=True
                    )

                # Clear pending state
                self.env["commandBuffer"]["voiceTestCompleted"] = False

                # Exit vmenu after successful application
                self.env["runtime"]["VmenuManager"].set_active(False)

            except Exception as e:
                # Revert on failure
                settings_manager.settings["speech"]["driver"] = old_driver
                settings_manager.settings["speech"]["module"] = old_module
                settings_manager.settings["speech"]["voice"] = old_voice

                # Try to reinitialize with old settings
                if "SpeechDriver" in self.env["runtime"]:
                    try:
                        SpeechDriver = self.env["runtime"]["SpeechDriver"]
                        SpeechDriver.shutdown()
                        SpeechDriver.initialize(self.env)
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "dynamicVoiceMenu: Error reinitializing speech driver: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )

                self.env["runtime"]["OutputManager"].present_text(
                    f"Failed to apply voice, reverted: {str(e)}",
                    interrupt=False,
                    flush=False,
                )

        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                f"Apply voice error: {str(e)}", interrupt=False, flush=False
            )

    def set_callback(self, callback):
        pass


def add_dynamic_voice_menus(VmenuManager):
    """Add dynamic voice menus to vmenu system"""
    try:
        env = VmenuManager.env

        # Get speech modules
        modules = get_speechd_modules()
        if not modules:
            return

        # Create voice browser submenu
        voice_browser_menu = {}

        # Add apply voice command
        apply_command = DynamicApplyVoiceCommand(env)
        voice_browser_menu["Apply Tested Voice Action"] = apply_command

        # Add modules as submenus
        for module in modules[
            :8
        ]:  # Limit to 8 modules to keep menu manageable
            module_menu = {}

            # Get voices for this module
            voices = get_module_voices(module)
            if voices:

                # Add voice commands
                for voice in voices:
                    voice_command = DynamicVoiceCommand(module, voice, env)
                    module_menu[f"{voice} Action"] = voice_command
            else:
                module_menu["No voices available Action"] = (
                    create_info_command(f"No voices found for {module}", env)
                )

            voice_browser_menu[f"{module} Menu"] = module_menu

        # Add to main menu dict
        VmenuManager.menuDict["Voice Browser Menu"] = voice_browser_menu

    except Exception as e:
        # Use debug manager instead of print for error logging
        if "DebugManager" in env["runtime"]:
            env["runtime"]["DebugManager"].write_debug_out(
                f"Error creating dynamic voice menus: {e}",
                debug.DebugLevel.ERROR,
            )
        else:
            print(f"Error creating dynamic voice menus: {e}")


def create_info_command(message, env):
    """Create a simple info command"""

    class InfoCommand:
        def __init__(self, message, env):
            self.message = message
            self.env = env

        def initialize(self, environment):
            pass

        def shutdown(self):
            pass

        def get_description(self):
            return self.message

        def run(self):
            self.env["runtime"]["OutputManager"].present_text(
                self.message, interrupt=True
            )

        def set_callback(self, callback):
            pass

    return InfoCommand(message, env)


def get_speechd_modules():
    """Get available speech modules"""
    try:
        result = subprocess.run(
            ["spd-say", "-O"], capture_output=True, text=True, timeout=5
        )
        if result.returncode == 0:
            lines = result.stdout.strip().split("\n")
            return [line.strip() for line in lines[1:] if line.strip()]
    except Exception:
        pass
    return []


def get_module_voices(module):
    """Get voices for a module"""
    try:
        result = subprocess.run(
            ["spd-say", "-o", module, "-L"],
            capture_output=True,
            text=True,
            timeout=8,
        )
        if result.returncode == 0:
            lines = result.stdout.strip().split("\n")
            voices = []
            for line in lines[1:]:
                if not line.strip():
                    continue
                if module.lower() == "espeak-ng":
                    voice = process_espeak_voice(line)
                    if voice:
                        voices.append(voice)
                else:
                    voices.append(line.strip())
            return voices
    except Exception:
        pass
    return []


def process_espeak_voice(voiceLine):
    """Process espeak voice format"""
    try:
        parts = [p for p in voiceLine.split() if p]
        if len(parts) < 2:
            return None
        lang_code = parts[-2].lower()
        variant = parts[-1].lower()
        return (
            f"{lang_code}+{variant}"
            if variant and variant != "none"
            else lang_code
        )
    except Exception:
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time

from fenrirscreenreader.core import debug


class MemoryManager:
    def __init__(self):
        self.listStorage = {}

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def list_storage_valid(self, name, checkIndex=False):
        try:
            if checkIndex:
                index = self.listStorage[name]["index"]
                if index == -1:
                    return self.listStorage[name]["list"] == []
                return self.listStorage[name]["list"][index] is not None
            else:
                return isinstance(self.listStorage[name]["list"], list)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "list_storage_valid " + str(e), debug.DebugLevel.ERROR
            )
        return False

    def add_value_to_first_index(self, name, value):
        if not self.list_storage_valid(name):
            return
        if self.listStorage[name]["maxLength"] is None:
            # Fallback: if maxLength is still None, apply default limit of 1000
            self.listStorage[name]["list"] = [value] + self.listStorage[name][
                "list"
            ][:999]
        else:
            self.listStorage[name]["list"] = [value] + self.listStorage[name][
                "list"
            ][: self.listStorage[name]["maxLength"] - 1]
        self.listStorage[name]["index"] = 0

    def add_index_list(self, name, maxLength=1000, currList=[], curr_index=-1):
        if len(currList) != 0 and (curr_index == -1):
            curr_index = 0
        self.listStorage[name] = {
            "list": currList,
            "index": curr_index,
            "maxLength": maxLength,
        }

    def is_last_index(self, name):
        if not self.list_storage_valid(name):
            return False
        return (
            self.listStorage[name]["index"]
            == len(self.listStorage[name]["list"]) - 1
        )

    def is_first_index(self, name):
        if not self.list_storage_valid(name):
            return False
        return self.listStorage[name]["index"] == 0

    def get_next_index(self, name):
        if not self.list_storage_valid(name):
            return False
        if self.is_index_list_empty(name):
            self.listStorage[name]["index"] = -1
            return False
        self.listStorage[name]["index"] += 1
        if (
            self.listStorage[name]["index"]
            > len(self.listStorage[name]["list"]) - 1
        ):
            self.listStorage[name]["index"] = 0
        return True

    def set_pref_index(self, name):
        if not self.list_storage_valid(name):
            return False
        if self.is_index_list_empty(name):
            self.listStorage[name]["index"] = -1
            return False
        self.listStorage[name]["index"] -= 1
        if self.listStorage[name]["index"] < 0:
            self.listStorage[name]["index"] = (
                len(self.listStorage[name]["list"]) - 1
            )
        return True

    def set_first_index(self, name):
        if not self.list_storage_valid(name):
            return False
        if self.is_index_list_empty(name):
            self.listStorage[name]["index"] = -1
            return False
        self.listStorage[name]["index"] = 0
        return True

    def get_index_list_len(self, name):
        if not self.list_storage_valid(name):
            return 0
        if self.is_index_list_empty(name):
            self.listStorage[name]["index"] = -1
            return 0
        return len(self.listStorage[name])

    def set_last_index(self, name):
        if not self.list_storage_valid(name):
            return False
        if self.is_index_list_empty(name):
            self.listStorage[name]["index"] = -1
            return False
        self.listStorage[name]["index"] = (
            len(self.listStorage[name]["list"]) - 1
        )
        return True

    def clear_current_index_list(self, name):
        if not self.list_storage_valid(name):
            return False
        self.listStorage[name]["list"] = []
        self.listStorage[name]["index"] = -1

    def get_current_index(self, name):
        if not self.list_storage_valid(name):
            return False
        if self.is_index_list_empty(name):
            self.listStorage[name]["index"] = -1
            return False
        try:
            return self.listStorage[name]["index"]
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "MemoryManager get_current_index: Error accessing index for "
                + name
                + ": "
                + str(e),
                debug.DebugLevel.ERROR,
            )
            return -1

    def is_index_list_empty(self, name):
        if not self.list_storage_valid(name):
            return False
        return len(self.listStorage[name]["list"]) == 0

    def get_index_list_element(self, name):
        if not self.list_storage_valid(name):
            return None
        if self.is_index_list_empty(name):
            self.listStorage[name]["index"] = -1
            return None
        curr_index = self.get_current_index(name)
        if curr_index == -1:
            return None
        try:
            return self.listStorage[name]["list"][curr_index]
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "MemoryManager get_index_list_element: Error accessing element for "
                + name
                + ": "
                + str(e),
                debug.DebugLevel.ERROR,
            )
            return None



#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "Save current session settings to configuration file"

    def run(self):
        self.env["runtime"]["OutputManager"].present_text(
            "Saving current session settings to configuration file...",
            interrupt=True,
        )

        try:
            # This calls the settings manager's save method which writes
            # current runtime settings to file
            self.env["runtime"]["SettingsManager"].save_settings()

            self.env["runtime"]["OutputManager"].present_text(
                "Session settings saved successfully!",
                interrupt=False,
                flush=False,
            )
            self.env["runtime"]["OutputManager"].present_text(
                "All temporary changes are now permanent.",
                interrupt=False,
                flush=False,
            )

        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                f"Failed to save settings: {str(e)}",
                interrupt=False,
                flush=False,
            )

    def set_callback(self, callback):
        pass
